//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef PARAMETERS_H
#define PARAMETERS_H

#include <cassert>
#include <QObject>

class Parameters : public QObject
{
    Q_OBJECT

public:
    Parameters() :
        pcaTol(Tolerance::MaxRelError),
        tpsTol(Tolerance::MaxRelError),
        tpsSmoothness(1.0),
        splTol(Tolerance::MaxRelError),
        splAlpha(0.5)
    {
        pcaTol.setRelError(2e-2);
        tpsTol.setRelError(4e-2);
        splTol.setRelError(6e-2);
    }

    class Tolerance
    {
    public:
        enum Type { MaxRelError, DegFreedom };

        Tolerance() {}
        Tolerance(Type type) :
            type(type)
        {
            switch (type)
            {
            case MaxRelError:
                value.maxRelError = 5e-2;
                break;
            case DegFreedom:
                value.degreesFreedom = 10;
                break;
            }
        }

        bool operator==(const Tolerance &t) const
        {
            return (type == t.type && type == MaxRelError && value.maxRelError == t.value.maxRelError) ||
                   (type == t.type && type == DegFreedom  && value.degreesFreedom == t.value.degreesFreedom);
        }
        bool operator!=(const Tolerance &t) const
        {
            return !(*this == t);
        }

        void setRelError(double e) { type = MaxRelError; value.maxRelError = e; }
        void setDegFreedom(int n) { type = DegFreedom; value.degreesFreedom = n; }
        Type getType() const { return type; }
        double getMaxRelError() const { assert(type == MaxRelError); return value.maxRelError;    }
        int getDegreesFreedom() const { assert(type == DegFreedom);  return value.degreesFreedom; }
    
    private:
        Type type;
        union TolValue {
            double maxRelError;
            int degreesFreedom;
        } value;
    };

    void setPCATol(Tolerance t)     { if (pcaTol != t)        { pcaTol        = t; emit pcaParamsChanged(); } }
    void setTPSTol(Tolerance t)     { if (tpsTol != t)        { tpsTol        = t; emit tpsParamsChanged(); } }
    void setTPSSmoothness(double s) { if (tpsSmoothness != s) { tpsSmoothness = s; emit tpsParamsChanged(); } }
    void setSPLTol(Tolerance t)     { if (splTol != t)        { splTol        = t; emit splParamsChanged(); } }
    void setSPLAlpha(double a)      { if (splAlpha != a)      { splAlpha      = a; emit splParamsChanged(); } }

    const Tolerance &PCATol() const { return pcaTol; }
    const Tolerance &TPSTol() const { return tpsTol; }
    double TPSSmoothness()    const { return tpsSmoothness; }
    const Tolerance &SPLTol() const { return splTol; }
    double SPLAlpha()         const { return splAlpha; }

signals:
    void pcaParamsChanged();
    void tpsParamsChanged();
    void splParamsChanged();

private:
    Tolerance pcaTol;
    Tolerance tpsTol;
    double    tpsSmoothness;
    Tolerance splTol;
    double    splAlpha;
};

#endif // PARAMETERS_H
