//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "CubicSplineRecon.h"
#include "ThinPlateRecon.h"
#include "Sequence.h"

#include <iostream>
#include <QElapsedTimer>
#include <QProgressDialog>
#include <Eigen/QR>
#include <Eigen/SVD>

CubicSplineRecon::CubicSplineRecon(const Eigen::MatrixXd &targetPoints,
                                   const Parameters::Tolerance &tol,
                                   Sequence &sequence,
                                   double paramPower,
                                   QProgressDialog *progress)
    : sequence(sequence), paramPower(paramPower), resampleFactor(1.0)
{
    switch (tol.getType())
    {
    case Parameters::Tolerance::DegFreedom:
        compute(targetPoints, qMin(tol.getDegreesFreedom(), sequence.getNumFrames() + 2));
        break;
    case Parameters::Tolerance::MaxRelError:
        for (int nPoints = 4; nPoints <= sequence.getNumFrames() + 2; ++nPoints)
        {
            double maxRelError = 0;
            compute(targetPoints, nPoints);
            if (progress) progress->setValue(progress->value() + 1);
            
            for (int frame = 0; frame < sequence.getNumFrames(); ++frame)
            {
                Eigen::VectorXd recon;
                QPointF pt = evaluate(frame);
                sequence.evaluateThinPlate(pt.x(), pt.y(), recon);
                double thisError = sequence.relErr(recon, frame);
                if (thisError > maxRelError) maxRelError = thisError;
            }

            if (maxRelError < tol.getMaxRelError()) break;
        }
        break;
    }
}

CubicSplineRecon::CubicSplineRecon(Sequence &sequence, QPointF position,
                                   int numControlPoints)
    : sequence(sequence), paramPower(0.0), resampleFactor(1.0)
{
    assert(numControlPoints >= 4);

    knots.push_back(0);
    knots.push_back(0);
    knots.push_back(0);
    for (int i = 1; i < numControlPoints - 3; ++i) knots.push_back(i);
    knots.push_back(numControlPoints - 3);
    knots.push_back(numControlPoints - 3);
    knots.push_back(numControlPoints - 3);

    controlPoints = Eigen::MatrixXd::Zero(numControlPoints, 2);
    controlPoints.col(0).setConstant(position.x());
    controlPoints.col(1).setConstant(position.y());

    numSamples = numControlPoints - 2;

    paramInterp.clear();
    for (int i = 0; i < numSamples; ++i)
    {
        params.push_back(i);
        paramInterp.addDataValue(i, i);
    }
    paramInterp.computeInterpolant();
}

void CubicSplineRecon::compute(const Eigen::MatrixXd &targetPoints, int numControlPoints)
{
    QElapsedTimer timer;
    timer.start();
    controlPoints = Eigen::MatrixXd::Zero(numControlPoints, targetPoints.cols());

    computeParams(paramPower, targetPoints);
    numSamples = sequence.getNumFrames();
    assert(numSamples == params.size());
    int degree        = 3;
    int polynomPieces = numControlPoints - degree;
    int numIntKnots   = polynomPieces - 1;
    int numKnots      = numIntKnots + 2 * degree;

    paramInterp.clear();
    for (int i = 0; i < numSamples; ++i)
    {
        paramInterp.addDataValue(i, params[i]);
    }
    paramInterp.computeInterpolant();

    knots.assign(numKnots, 0);
    for (int i = 0; i < degree; ++i)
    {
        // knots[i] = 0;
        knots[numKnots - i - 1] = params.back();
    }
    for (int i = 0; i < numIntKnots; ++i)
    {
        if ((i + 1) * (numSamples - 1) % (numIntKnots + 1) == 0)
        {
            knots[degree + i] = params[(i + 1) * (numSamples - 1) / (numIntKnots + 1)];
        }
        else
        {
            int firstKnot = (i + 1) * (numSamples - 1) / (numIntKnots + 1);
            double blend  = (i + 1) * (numSamples - 1) / static_cast<double>(numIntKnots + 1);
            blend -= firstKnot;

            knots[degree + i] =      blend  * params[firstKnot + 1] +
                                (1 - blend) * params[firstKnot];
        }
    }
    
    Eigen::VectorXd temp(numControlPoints);
    Eigen::MatrixXd splineCoeffs = Eigen::MatrixXd::Zero(numSamples, numControlPoints);
    for (int i = 0; i < numSamples; ++i)
    {
        directEvaluate(params[i], temp, 0, true);
        splineCoeffs.row(i) = temp;
    }

    // No end conditions
    // Eigen::LDLT<Eigen::MatrixXd> ldlt(splineCoeffs.transpose() * splineCoeffs);
    // controlPoints = ldlt.solve(splineCoeffs.transpose() * targetPoints);
    
    // End conditions
    Eigen::MatrixXd constraints = Eigen::MatrixXd::Zero(2, numControlPoints);

    // Natural end conditions
    // directEvaluate(knots[0], temp, 2, true);
    // constraints.row(0) = temp;
    // directEvaluate(knots[numKnots - 1], temp, 2, true);
    // constraints.row(1) = temp;

    // Not-a-knot end conditions
    directEvaluate(knots[3] / 2.0, temp, 3, true);
    constraints.row(0) = temp;
    directEvaluate((knots[3] + knots[4]) / 2.0, temp, 3, true);
    constraints.row(0) -= temp;
    directEvaluate((knots[numKnots - 3] + knots[numKnots - 4]) / 2.0, temp, 3, true);
    constraints.row(1) = temp;
    directEvaluate((knots[numKnots - 4] + knots[numKnots - 5]) / 2.0, temp, 3, true);
    constraints.row(1) -= temp;

    Eigen::HouseholderQR<Eigen::MatrixXd> cQR(constraints.transpose());
    Eigen::MatrixXd AQ = splineCoeffs;
    AQ.applyOnTheRight(cQR.householderQ());
    
    controlPoints.bottomRows(numControlPoints - 2) = AQ.rightCols(numControlPoints - 2)
                                                       .jacobiSvd(Eigen::ComputeThinU | Eigen::ComputeThinV)
                                                       .solve(targetPoints);
    controlPoints.applyOnTheLeft(cQR.householderQ());
    
    std::cout << "Calculating SPL took " << timer.elapsed() << "ms" << std::endl;
}

void CubicSplineRecon::directEvaluate(double t, Eigen::VectorXd &result, int nDerivs, bool onIdentity) const
{
    static const int degree = 3;
    if (onIdentity)
    {
        assert(result.size() == knots.size() - degree + 1);
        result.setZero();
    }
    else
    {
        assert(result.size() == controlPoints.cols());
    }
    
    int knot_greater = find_if(knots.begin(), knots.end(), std::bind2nd(std::greater_equal<double>(), t)) - knots.begin();
    int polyPiece    = knot_greater - degree;
    if (polyPiece < 0) polyPiece = 0;
    if (polyPiece >= controlPoints.rows() - degree) polyPiece = controlPoints.rows() - degree - 1;

    Eigen::MatrixXd points;
    if (onIdentity)
    {
        points = Eigen::MatrixXd::Identity(degree + 1, degree + 1);
    }
    else
    {
        points = controlPoints.block(polyPiece, 0, degree + 1, controlPoints.cols());
    }
    for (int d = degree; d > degree - nDerivs; --d)
    {
        Eigen::MatrixXd divDiffs = Eigen::MatrixXd::Zero(d, d + 1);

        for (int i = degree - d; i < degree; ++i)
        {
            int fix = i - degree + d;
            double denom = (knots[polyPiece + i + d] - knots[polyPiece + i]);
            divDiffs(fix, fix) = -d / denom;
            divDiffs(fix, fix + 1) = d / denom;
        }

        points = divDiffs * points;
    }
    for (int d = degree - nDerivs; d > 0; --d)
    {
        Eigen::MatrixXd filter = Eigen::MatrixXd::Zero(d, d + 1);

        for (int i = degree - d; i < degree; ++i)
        {
            int fix = i - degree + d;
            double blend = (t - knots[polyPiece + i]) / 
                           (knots[polyPiece + i + d] - knots[polyPiece + i]);
            filter(fix, fix) = 1 - blend;
            filter(fix, fix + 1) = blend;
        }

        points = filter * points;
    }

    if (onIdentity)
    {
        result.segment(polyPiece, degree + 1) = points.transpose();
    }
    else
    {
        result = points.transpose();
    }
}

QPointF CubicSplineRecon::directEvaluate(double t, int nDerivs) const
{
    Eigen::VectorXd result = Eigen::VectorXd::Zero(2);
    directEvaluate(t, result, nDerivs);
    return QPointF(result(0), result(1));
}

void CubicSplineRecon::evaluate(double t, Eigen::VectorXd &result, int nDerivs) const
{
    assert(nDerivs <= 1);
    directEvaluate(paramInterp.evaluateInterpolant(resampleFactor * t), result, nDerivs);
    if (nDerivs == 1) result *= paramInterp.evaluateDerivative(resampleFactor * t);
}

QPointF CubicSplineRecon::evaluate(double t, int nDerivs) const
{
    Eigen::VectorXd result = Eigen::VectorXd::Zero(2);
    evaluate(t, result, nDerivs);
    return QPointF(result(0), result(1));
}

void CubicSplineRecon::setControlPoint(int pointNumber, QPointF newPoint)
{
    controlPoints(pointNumber, 0) = newPoint.x();
    controlPoints(pointNumber, 1) = newPoint.y();
    sequence.redrawSplineCurve();
}

void CubicSplineRecon::uniformResample(int nFrames)
{
    resampleFactor *= static_cast<double>(numSamples - 1) / (nFrames - 1);
    numSamples = nFrames;
}

void CubicSplineRecon::addFrozenFrames(int afterFrame, int nFrames, double newFrameVal)
{
    double newVal = paramInterp.evaluateInterpolant(resampleFactor * newFrameVal);
    paramInterp.addFrozenFrames(resampleFactor * afterFrame, nFrames, newVal);
    paramInterp.computeInterpolant();
    numSamples += (nFrames / resampleFactor);
}

void CubicSplineRecon::duplicateFrames(int fromFrame, int toFrame, int repetitions)
{
    int addFrames = repetitions * (toFrame - fromFrame);
    numSamples += addFrames;
    double paramFrom = params[fromFrame];
    double paramTo   = params[toFrame];

    for (unsigned i = toFrame; i < params.size(); ++i)
    {
        params[i] += repetitions * (paramTo - paramFrom);
    }
    params.insert(params.begin() + toFrame, addFrames, 0);
    for (int i = toFrame; i < toFrame + addFrames; ++i)
    {
        params[i] = paramTo + params[i - toFrame + fromFrame] - paramFrom;
    }
    paramInterp.clear();
    for (int i = 0; i < numSamples; ++i)
    {
        paramInterp.addDataValue(i, params[i]);
    }
    paramInterp.computeInterpolant();

    int minKnotIx = -1, maxKnotIx = -1;
    for (unsigned i = 0; i < knots.size(); ++i)
    {
        if (knots[i] > paramFrom && minKnotIx == -1) minKnotIx = i;
        if (knots[i] < paramTo) maxKnotIx = i;
    }
    ++maxKnotIx;
    int addPoints = repetitions * (maxKnotIx - minKnotIx);
    controlPoints.conservativeResize(controlPoints.rows() + addPoints, 2);
    controlPoints.block(minKnotIx + addPoints, 0, controlPoints.rows() - minKnotIx - addPoints, 2) =
        controlPoints.block(minKnotIx, 0, controlPoints.rows() - minKnotIx - addPoints, 2).eval();
    for (int i = 1; i < repetitions; ++i)
    {
        controlPoints.block(minKnotIx + i * (maxKnotIx - minKnotIx), 0, maxKnotIx - minKnotIx, 2) =
            controlPoints.block(minKnotIx, 0, maxKnotIx - minKnotIx, 2);
    }

    for (unsigned i = maxKnotIx; i < knots.size(); ++i)
    {
        knots[i] += repetitions * (paramTo - paramFrom);
    }
    knots.insert(knots.begin() + maxKnotIx, addPoints, 0);
    for (int i = maxKnotIx; i < maxKnotIx + addPoints; ++i)
    {
        knots[i] = paramTo + knots[i - maxKnotIx + minKnotIx] - paramFrom;
    }
}

void CubicSplineRecon::computeParams(double alpha, const Eigen::MatrixXd &targetPoints)
{
    params.clear();
    params.push_back(0);

    for (int i = 1; i < targetPoints.rows(); ++i)
    {
        params.push_back(params.back() + pow((targetPoints.row(i) - targetPoints.row(i - 1)).norm(), alpha));
    }
}
