//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "EmbedderMDS.h"
#include "Sequence.h"
#include "MovableEllipse.h"

#include <QApplication>
#include <QGraphicsScene>

EmbedderMDS::EmbedderMDS(const Eigen::MatrixXd &normalizedFeatures,
                         Sequence &sequence) :
    normalizedFeatures(normalizedFeatures), sequence(sequence), iterating(false)
{
    int frames = sequence.getNumFrames();
    embedding = Eigen::MatrixXd::Zero(frames, 2);
    realDists = Eigen::MatrixXd::Zero(frames, frames);

    for (int f = 0; f < frames; ++f)
    {
        embedding(f, 0) = sequence.getSubspace().getCoeff(f, 0);
        embedding(f, 1) = sequence.getSubspace().getCoeff(f, 1);
    }
    
    for (int i = 0; i < frames - 1; ++i)
    {
        for (int j = i + 1; j < frames; ++j)
        {
            calcRealDist(i, j);
        }
    }

    points.resize(frames, NULL);
}

double EmbedderMDS::frameX(int frame) const
{
    return embedding(frame, 0);
}

double EmbedderMDS::frameY(int frame) const
{
    return embedding(frame, 1);
}

QRectF EmbedderMDS::boundingRect() const
{
    double minX = embedding.col(0).minCoeff();
    double minY = embedding.col(1).minCoeff();
    return QRectF(minX, minY,
                  embedding.col(0).maxCoeff() - minX,
                  embedding.col(1).maxCoeff() - minY);
}

void EmbedderMDS::setPoint(int frame, MovableEmbedderEllipse *point)
{
    points[frame] = point;
    point->setX(frameX(frame));
    point->setY(frameY(frame));
    point->setEmbedder(this);
}

void EmbedderMDS::unpin(int frame)
{
    if (points[frame] == NULL) return;
    points[frame]->unpin();
    iterateToConvergence();
}

void EmbedderMDS::unpinSelected()
{
    if (iteratingNow()) return;

    for (int frame = 0; frame < sequence.getNumFrames(); ++frame)
    {
        if (points[frame] == NULL) continue;
        if (points[frame]->isSelected()) points[frame]->unpin();
    }

    iterateToConvergence();
}

void EmbedderMDS::unpinAll()
{
    if (iteratingNow()) return;

    for (int frame = 0; frame < sequence.getNumFrames(); ++frame)
    {
        if (points[frame] == NULL) continue;
        points[frame]->unpin();
    }

    iterateToConvergence();
}

void EmbedderMDS::pointUpdated(MovableEmbedderEllipse * /* point */)
{
    iterateToConvergence();
}

void EmbedderMDS::iterateToConvergence()
{
    iterating = true;
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    double oldStress, newStress = stress();
    
    for (int frame = 0; frame < sequence.getNumFrames(); ++frame)
    {
        if (points[frame] == NULL) continue;
        points[frame]->setFlag(QGraphicsItem::ItemSendsGeometryChanges, false);
    }

    int frames = sequence.getNumFrames();
    Eigen::MatrixXd B(frames, frames);

    do {
        oldStress = newStress;

        B.setZero();
        for (int i = 0; i < frames - 1; ++i)
        {
            for (int j = i + 1; j < frames; ++j)
            {
                double delta    = embedDist(i, j);
                double weighted = (delta == 0 ? 0 : realDists(i, j) / delta);

                B(i, i) += weighted;
                B(j, j) += weighted;
                B(i, j) -= weighted;
                B(j, i) -= weighted;
            }
        }
        embedding = B * embedding / frames;
        embeddingChanged();

        syncWithPoints();
        newStress = stress();
    } while ((oldStress / newStress) > 1 + 1e-7);
    
    for (int frame = 0; frame < sequence.getNumFrames(); ++frame)
    {
        if (points[frame] == NULL) continue;
        points[frame]->setFlag(QGraphicsItem::ItemSendsGeometryChanges, true);
    }
    QApplication::restoreOverrideCursor();
    iterating = false;
}

bool EmbedderMDS::iteratingNow() const
{
    return iterating;
}

void EmbedderMDS::syncWithPoints()
{
    for (int frame = 0; frame < sequence.getNumFrames(); ++frame)
    {
        if (points[frame] == NULL) continue;

        if (points[frame]->isPinned())
        {
            embedding(frame, 0) = points[frame]->x();
            embedding(frame, 1) = points[frame]->y();
        }
        else
        {
            points[frame]->setX(embedding(frame, 0));
            points[frame]->setY(embedding(frame, 1));
        }

        if (frame > 0)
        {
            QGraphicsLineItem *line = points[frame]->getPrevLine();
            if (line == NULL) continue;

            line->setLine(embedding(frame - 1, 0), embedding(frame - 1, 1),
                          embedding(frame, 0), embedding(frame, 1));
        }
    }

    QApplication::processEvents();
}

void EmbedderMDS::calcRealDist(int i, int j)
{
    realDists(i, j) = (normalizedFeatures.col(i) - normalizedFeatures.col(j)).norm();
}

double EmbedderMDS::embedDist(int i, int j) const
{
    return (embedding.row(i) - embedding.row(j)).norm();
}

double EmbedderMDS::stress() const
{
    double stress = 0;
    for (int i = 0; i < sequence.getNumFrames() - 1; ++i)
    {
        for (int j = i + 1; j < sequence.getNumFrames(); ++j)
        {
            double diff = (realDists(i, j) - embedDist(i, j));
            stress += diff * diff;
        }
    }
    return stress;
}
