//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "GraphicsView.h"
#include <QGraphicsEllipseItem>
#include <QMouseEvent>

GraphicsView::GraphicsView(QWidget *parent) :
    QGraphicsView(parent),
    currentPoint(NULL),
    aspectRatioMode(Qt::KeepAspectRatio),
    mouseEventsFlag(true),
    currentPointVisibleFlag(true)
{
}

void GraphicsView::resizeToFit()
{
	if (scene() != NULL)
	{
        QRectF boundingRect = scene()->sceneRect();
		fitInView(boundingRect, aspectRatioMode);
        if (currentPoint == NULL)
        {
            double range = qMax(boundingRect.width(), boundingRect.height());
            double radius = range / 120.0;
            currentPoint = scene()->addEllipse(-radius / 2.0, -radius / 2.0,
                                               radius, radius,
                                               QPen(Qt::red), QBrush(Qt::red));
            currentPoint->setVisible(currentPointVisibleFlag);
            currentPoint->setZValue(2);
            currentPoint->setX(boundingRect.center().x());
            currentPoint->setY(boundingRect.center().y());
        }
	}
}

void GraphicsView::resizeEvent(QResizeEvent *event)
{
    QGraphicsView::resizeEvent(event);
	resizeToFit();
}

void GraphicsView::mousePressEvent(QMouseEvent *event)
{
    if (scene() != NULL && mouseEventsFlag)
	{
        setCurrentPoint(mapToScene(event->pos()));
    }
    else QGraphicsView::mousePressEvent(event);
}

void GraphicsView::mouseMoveEvent(QMouseEvent *event)
{
    if (scene() != NULL && mouseEventsFlag)
	{
        setCurrentPoint(mapToScene(event->pos()));
    }
    else QGraphicsView::mouseMoveEvent(event);
}

void GraphicsView::setCurrentPoint(QPointF pos)
{
    currentPoint->setX(pos.x());
    currentPoint->setY(pos.y());
    emit newPointPosition(pos.x(), pos.y());
}

QPointF GraphicsView::getCurrentPoint() const
{
    QPointF res;
    res.setX(currentPoint->x());
    res.setY(currentPoint->y());
    return res;
}

void GraphicsView::setScene(QGraphicsScene *newScene)
{
    if (scene() != newScene) detachFromScene();
    QGraphicsView::setScene(newScene);
    if (NULL != newScene)
    {
        newScene->setItemIndexMethod(QGraphicsScene::NoIndex);
    }
    resizeToFit();
}

void GraphicsView::setCurrentPointVisible(bool visible)
{
    currentPointVisibleFlag = visible;
    if (currentPoint != NULL)
    {
        currentPoint->setVisible(visible);
    }
}

void GraphicsView::detachFromScene()
{
    if (scene() != NULL && currentPoint != NULL)
    {
        scene()->removeItem(currentPoint);
        delete currentPoint;
        currentPoint = NULL;
    }
}
