//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "ui_parametersdialog.h"
#include "ui_sequencedialog.h"

#include <QDir>
#include <QFile>
#include <QFileDialog>
#include <QInputDialog>
#include <QGraphicsPathItem>
#include <QProgressDialog>
#include <QStringBuilder>
#include <QTimer>

#include "DuplicateFrameDialog.h"
#include "EmbedderMDS.h"
#include "OpenMeshDrawer.h"
#include "Reps.h"
#include "ResampleDialog.h"
#include "Sequence.h"
#include "ThinPlateRecon.h"

MainWindow *MainWindow::instance = NULL;

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    playing(false),
    statusLabelsAdded(false),
    currentRep(Vertex),
    vConverter(NULL), dConverter(NULL), lConverter(NULL)
{
    QGLFormat format;
    format.setSampleBuffers(true);
    QGLFormat::setDefaultFormat(format);

    ui->setupUi(this);
    drawer = new OpenMeshDrawer(*ui->viewer);

    animTimer = new QTimer(this);
    connect(animTimer, SIGNAL(timeout()), this, SLOT(advanceFrame()));
    connect(ui->graphicsView, SIGNAL(newPointPosition(double, double)),
            this,               SLOT(evaluateThinPlate(double, double)));
    connect(ui->timeSlider,   SIGNAL(valueChanged(double)),
            this,               SLOT(evaluateSpline(double)));

    on_tabs_currentChanged(-1);
    int splitterHalf = ui->splitter->width() / 2;
    ui->splitter->setSizes(QList<int>() << splitterHalf << splitterHalf);
}

MainWindow::~MainWindow()
{
    if (NULL != vConverter) delete vConverter;
    if (NULL != dConverter) delete dConverter;
    if (NULL != lConverter) delete lConverter;

    for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
         it != sequences.end(); ++it)
    {
        delete it->second;
    }
    delete drawer;
    delete ui;
}

void MainWindow::evaluateSpline(double t)
{
    if (ui->showSourceCheckBox->isChecked())
    {
        Eigen::VectorXd reconVec;
        int frame = static_cast<QSlider *>(ui->timeSlider)->value();
        frame *= sequences[currentRep]->getSpline().getResampleFactor();
        frame = qMin(sequences[currentRep]->getNumFrames() - 1, frame);
        sequences[currentRep]->getFrame(frame, reconVec);
        setMeshFromVector(reconVec);
    }

    ui->graphicsView->setCurrentPoint(sequences[currentRep]->getSpline().evaluate(t));
}

void MainWindow::evaluateThinPlate(double x, double y)
{
    if (ui->snapToCurveCheckBox->isChecked() &&
        ui->showSourceCheckBox->isChecked()) return;

    Eigen::VectorXd reconVec;
    sequences[currentRep]->evaluateThinPlate(x, y, reconVec);
    setMeshFromVector(reconVec);
}

void MainWindow::setMeshFromVector(Eigen::VectorXd &vec)
{
    switch (currentRep)
    {
    case Vertex:
        {
            VertexRep v;
            v.setFromVector(vec);
            vConverter->setMeshFromRep(target, v);
        }
        break;
    case DefGrad:
        {
            DefGradRep d;
            d.setFromVector(vec);
            dConverter->setMeshFromRep(target, d);
        }
        break;
    case LinRotInv:
        {
            LinRotInvRep l;
            l.setFromVector(vec);
            lConverter->setMeshFromRep(target, l);
        }
        break;
    }

    drawer->meshChanged();
}

void MainWindow::setFitSpline(bool fit)
{
    ui->actionFitSpline->setChecked(fit);
}

void MainWindow::splineCurveChanged()
{
    ui->graphicsView->resizeToFit();
    if (ui->tabs->currentWidget() == ui->keyframeTab) {
        on_keyframeSpinBox_valueChanged(ui->keyframeSpinBox->value());
    } else if (ui->snapToCurveCheckBox->isChecked()) {
        evaluateSpline(ui->timeSlider->value());
    } else if (NULL != ui->graphicsView->scene()) {
        QPointF point = ui->graphicsView->getCurrentPoint();
        evaluateThinPlate(point.x(), point.y());
    }
}

void MainWindow::on_actionLoadSequence_triggered()
{
    Ui::SequenceDialog sequenceDialogUi;
    QDialog sequenceDialog(this);
    sequenceDialogUi.setupUi(&sequenceDialog);
    connect(sequenceDialogUi.fileNameRootButton, SIGNAL(updateFileName(QString)),
            sequenceDialogUi.fileNameRoot, SLOT(setText(QString)));
    sequenceDialogUi.fileNameRoot->setText(QDir::currentPath() + "/");
    sequenceDialogUi.toFrame->setValue(99);

    if (sequenceDialog.exec() == QDialog::Accepted)
    {
        int fromFrame    = sequenceDialogUi.fromFrame->value();
        int toFrame      = sequenceDialogUi.toFrame->value();

        ui->snapToCurveCheckBox->setChecked(false);
        ui->graphicsView->setScene(NULL);
        for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
             it != sequences.end(); ++it)
        {
            delete it->second;
        }
        sequences.clear();
        loadSequence(sequenceDialogUi.fileNameRoot->text(), fromFrame, toFrame,
                     sequenceDialogUi.numberWidth->value(), sequenceDialogUi.extension->text());
    }
}

void MainWindow::on_actionParameters_triggered()
{
    Ui::ParametersDialog parametersDialogUi;
    QDialog parametersDialog(this);
    parametersDialogUi.setupUi(&parametersDialog);
    setDialogFromParameters(parametersDialogUi);

    if (sequences.find(currentRep) != sequences.end() && !sequences[currentRep]->hasFeatureMatrix())
    {
        parametersDialogUi.pcaRelErrRadio->setEnabled(false);
        parametersDialogUi.tpsRelErrRadio->setEnabled(false);
        parametersDialogUi.splRelErrRadio->setEnabled(false);
    }

    if (parametersDialog.exec() == QDialog::Accepted)
    {
        setParametersFromDialog(parametersDialogUi);
        compressCurrentRep();
    }
}

void MainWindow::on_actionVertex_triggered()
{
    if (currentRep != Vertex)
    {
        currentRep = Vertex;
        compressCurrentRep();
        splineCurveChanged();
    }
    setCheckedRep();
}

void MainWindow::on_actionDefGrad_triggered()
{
    if (currentRep != DefGrad)
    {
        currentRep = DefGrad;
        compressCurrentRep();
        splineCurveChanged();
    }
    setCheckedRep();
}

void MainWindow::on_actionLinRotInv_triggered()
{
    if (currentRep != LinRotInv)
    {
        currentRep = LinRotInv;
        compressCurrentRep();
        splineCurveChanged();
    }
    setCheckedRep();
}

void MainWindow::on_actionReconstructFromShapes_triggered()
{
    if (sequences.find(currentRep) == sequences.end()) return;

    Ui::SequenceDialog sequenceDialogUi;
    QDialog sequenceDialog(this);
    sequenceDialogUi.setupUi(&sequenceDialog);
    connect(sequenceDialogUi.fileNameRootButton, SIGNAL(updateFileName(QString)),
            sequenceDialogUi.fileNameRoot, SLOT(setText(QString)));
    sequenceDialogUi.fileNameRoot->setText(QDir::currentPath() + "/");
    sequenceDialogUi.fromFrame->setValue(0);
    sequenceDialogUi.toFrame->setValue(sequences[currentRep]->getPCAComps());

    if (sequenceDialog.exec() == QDialog::Accepted)
    {
        int numComps = sequenceDialogUi.toFrame->value() - sequenceDialogUi.fromFrame->value();
        Mesh keyframe;
        bool errorCondition = false;

        for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
             it != sequences.end(); ++it)
        {
            it->second->initLSRecon(numComps);
        }

        for (int i = sequenceDialogUi.fromFrame->value(); i <= sequenceDialogUi.toFrame->value(); ++i)
        {
            QString fileName = QString("%1%2%3").arg(sequenceDialogUi.fileNameRoot->text())
                                                .arg(QString::number(i), sequenceDialogUi.numberWidth->value(), '0')
                                                .arg(sequenceDialogUi.extension->text());

            if (!QFile::exists(fileName))
            {
                errorCondition = true;
                break;
            }
            
            OpenMesh::IO::read_mesh(keyframe, qPrintable(fileName));

            VertexRep    vRep(*vConverter, keyframe);
            DefGradRep   dRep(*dConverter, keyframe);
            LinRotInvRep lRep(*lConverter, keyframe);

            sequences[Vertex]->addLSVector(i - sequenceDialogUi.fromFrame->value(), vRep.getVector());
            sequences[DefGrad]->addLSVector(i - sequenceDialogUi.fromFrame->value(), dRep.getVector());
            sequences[LinRotInv]->addLSVector(i - sequenceDialogUi.fromFrame->value(), lRep.getVector());
        }

        if (!errorCondition)
        {
            for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
                 it != sequences.end(); ++it)
            {
                it->second->switchToLSRecon();
            }
            compressCurrentRep();
        }
    }
}

void MainWindow::on_actionSwitchSubspace_triggered()
{
    if (sequences.find(currentRep) == sequences.end()) return;
    if (!sequences[currentRep]->usingLeastSquares()) return;

    int fromFrame = 0;
    int toFrame = sequences[currentRep]->getPCAComps();

    Ui::SequenceDialog sequenceDialogUi;
    QDialog sequenceDialog(this);
    sequenceDialogUi.setupUi(&sequenceDialog);
    connect(sequenceDialogUi.fileNameRootButton, SIGNAL(updateFileName(QString)),
            sequenceDialogUi.fileNameRoot, SLOT(setText(QString)));
    sequenceDialogUi.fileNameRoot->setText(QDir::currentPath() + "/");
    sequenceDialogUi.fromFrame->setValue(fromFrame);
    sequenceDialogUi.fromFrame->setEnabled(false);
    sequenceDialogUi.toFrame->setValue(toFrame);
    sequenceDialogUi.toFrame->setEnabled(false);

    if (sequenceDialog.exec() == QDialog::Accepted)
    {
        Parameters::Tolerance pcaTol;
        pcaTol.setDegFreedom(sequences[currentRep]->getPCAComps());
        parameters.setPCATol(pcaTol);

        Parameters::Tolerance tpsTol;
        tpsTol.setDegFreedom(sequences[currentRep]->getTPSSites());
        parameters.setTPSTol(tpsTol);

        Parameters::Tolerance splTol;
        splTol.setDegFreedom(sequences[currentRep]->getNumSplinePoints());
        parameters.setSPLTol(splTol);

        ui->showSourceCheckBox->setChecked(false);
        ui->showSourceCheckBox->setEnabled(false);

        // Can't compute thin plate spline when we don't have a feature matrix
        // to compare to. So make sure all sequences have it calculated first.
        for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
             it != sequences.end(); ++it)
        {
            it->second->computeAll();
        }

        loadSequence(sequenceDialogUi.fileNameRoot->text(), fromFrame, toFrame,
                     sequenceDialogUi.numberWidth->value(), sequenceDialogUi.extension->text(), false);
    }
}

void MainWindow::on_actionViewPoseSpaceConnections_toggled(bool checked)
{
    if (checked) ui->actionViewPoseSpacePositions->setChecked(true);

    for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
         it != sequences.end(); ++it)
    {
        it->second->viewPoseSpaceConnections(checked);
    }
}

void MainWindow::on_actionViewPoseSpacePositions_toggled(bool checked)
{
    if (!checked) ui->actionViewPoseSpaceConnections->setChecked(false);

    for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
         it != sequences.end(); ++it)
    {
        it->second->viewPoseSpacePositions(checked);
    }
}

void MainWindow::on_actionFitSpline_toggled(bool checked)
{
    for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
         it != sequences.end(); ++it)
    {
        it->second->fitSplineToData(checked);
    }
    compressCurrentRep();
}

void MainWindow::on_actionCreateUniformSpline_triggered()
{
    if (sequences.find(currentRep) == sequences.end()) return;

    bool ok;
    int controlPoints = QInputDialog::getInt(this, tr("Uniform spline"),
                                             tr("Control points:"), 4, 4, 1000, 1, &ok);

    if (ok)
    {
        ui->actionFitSpline->setChecked(false);
        sequences[currentRep]->uniformSplineCurve(controlPoints);
        compressCurrentRep();
    }
}

void MainWindow::on_actionUniformResample_triggered()
{
    if (sequences.find(currentRep) == sequences.end()) return;
    
    bool ok;
    int frames = QInputDialog::getInt(this, tr("Uniform resample"),
                                            tr("Number of frames:"), toFrame - fromFrame + 1, 2, 1000, 1, &ok);

    if (ok)
    {
        if (ui->timeSlider->value() > frames - 1)
        {
            ui->timeSlider->setValue(0);
        }

        ui->actionFitSpline->setChecked(false);
        sequences[currentRep]->uniformResample(frames);
        compressCurrentRep();
    }
}

void MainWindow::on_actionNonuniformResample_triggered()
{
    if (sequences.find(currentRep) == sequences.end()) return;

    double currentFrame = -1;
    if (ui->tabs->currentWidget() == ui->timeSliderTab && ui->snapToCurveCheckBox->isChecked())
    {
        currentFrame = ui->timeSlider->value();
    }
    ResampleDialog resampleDialog(sequences[currentRep]->getSpline(), this, currentFrame);
    resampleDialog.setGeometry(geometry().adjusted(0, 20, 0, 18));

    if (resampleDialog.exec() == QDialog::Accepted)
    {
        ui->actionFitSpline->setChecked(false);
        splineCurveChanged();
    }
}

void MainWindow::on_actionAddFrozenFrames_triggered()
{
    if (sequences.find(currentRep) == sequences.end()) return;

    bool ok;
    int frames = QInputDialog::getInt(this, tr("Add frozen frames"),
                                            tr("Number of frames:"), 1, 1, 1000, 1, &ok);

    if (ok)
    {
        ui->actionFitSpline->setChecked(false);
        sequences[currentRep]->addFrozenFrames(static_cast<int>(ui->timeSlider->value()), frames, ui->timeSlider->value());
        compressCurrentRep();
    }
}

void MainWindow::on_actionDuplicateFrames_triggered()
{
    if (sequences.find(currentRep) == sequences.end()) return;

    DuplicateFrameDialog duplicateFrameDialog(ui->timeSlider->value(), sequences[currentRep]->getSpline().maxOfDomain(), this);

    if (duplicateFrameDialog.exec() == QDialog::Accepted)
    {
        ui->actionFitSpline->setChecked(false);
        sequences[currentRep]->duplicateFrames(duplicateFrameDialog.fromFrame(),
                                               duplicateFrameDialog.toFrame(),
                                               duplicateFrameDialog.repetitions());
        compressCurrentRep();
    }
}

void MainWindow::on_snapToCurveCheckBox_toggled(bool checked)
{
    stopPlaying();

    if (sequences.find(currentRep) == sequences.end() ||
        sequences[currentRep]->hasFeatureMatrix())
    {
        ui->showSourceCheckBox->setEnabled(checked);
    }
    else
    {
        ui->showSourceCheckBox->setEnabled(false);
    }
    ui->playStopButton->setEnabled(checked);
    ui->timeSlider->setEnabled(checked);
    ui->graphicsView->setMouseEventsEnabled(!checked);

    if (checked)
    {
        ui->statusBar->showMessage("Frame " % QString::number(ui->timeSlider->value(), 'f', 2), 2000);
        evaluateSpline(ui->timeSlider->value());
    }
}

void MainWindow::on_showSourceCheckBox_toggled(bool checked)
{
    ui->timeSlider->setSnapToIntegers(checked);
    evaluateSpline(ui->timeSlider->value());
}

void MainWindow::on_playStopButton_clicked(void)
{
    if (playing)
    {
        stopPlaying();
    }
    else
    {
        startPlaying();
    }
}

void MainWindow::on_timeSlider_valueChanged(double value)
{
    ui->statusBar->showMessage("Frame " % QString::number(value, 'f', 2), 2000);
    if (value < ui->timeSlider->maximum())
    {
        ui->playStopButton->setEnabled(true);
    }
    else
    {
        ui->playStopButton->setEnabled(false);
    }
}

void MainWindow::on_tabs_currentChanged(int /* index */)
{
    bool setSplineVisible, setSplineControlVisible, setPosePointsEnabled, recalculate;
    if (sequences.find(currentRep) != sequences.end())
    {
        recalculate = sequences[currentRep]->tpsIsDirty();
    }

    if (ui->tabs->currentWidget() == ui->defineMapTab)
    {
        if (playing) stopPlaying();
        ui->graphicsView->setDragMode(QGraphicsView::RubberBandDrag);
        ui->graphicsView->setMouseEventsEnabled(false);
        ui->graphicsView->setCurrentPointVisible(false);
        drawer->setVisible(false);

        ui->actionViewPoseSpaceConnections->setChecked(true);
        ui->actionViewPoseSpacePositions->setChecked(true);
        setSplineVisible = false;
        setSplineControlVisible = false;
        setPosePointsEnabled = true;
        recalculate = false;
    }
    else if (ui->tabs->currentWidget() == ui->timeSliderTab)
    {
        ui->graphicsView->setCurrentPointVisible(true);
        ui->graphicsView->setDragMode(QGraphicsView::NoDrag);
        ui->graphicsView->setMouseEventsEnabled(!ui->snapToCurveCheckBox->isChecked());
        if (sequences.find(currentRep) != sequences.end())
        {
            if (ui->snapToCurveCheckBox->isChecked())
            {
                evaluateSpline(ui->timeSlider->value());
            }
            else
            {
                QPointF currentPoint = ui->graphicsView->getCurrentPoint();
                evaluateThinPlate(currentPoint.x(), currentPoint.y());
            }
        }
        drawer->setVisible(true);

        ui->actionViewPoseSpaceConnections->setChecked(false);
        ui->actionViewPoseSpacePositions->setChecked(false);
        setSplineVisible = true;
        setSplineControlVisible = false;
        setPosePointsEnabled = false;
    }
    else if (ui->tabs->currentWidget() == ui->editCurveTab)
    {
        if (playing) stopPlaying();
        ui->graphicsView->setDragMode(QGraphicsView::RubberBandDrag);
        ui->graphicsView->setMouseEventsEnabled(false);
        ui->graphicsView->setCurrentPointVisible(false);
        drawer->setVisible(false);

        ui->actionViewPoseSpaceConnections->setChecked(false);
        ui->actionViewPoseSpacePositions->setChecked(false);
        setSplineVisible = true;
        setSplineControlVisible = true;
        setPosePointsEnabled = false;
    }
    else if (ui->tabs->currentWidget() == ui->keyframeTab)
    {
        if (playing) stopPlaying();
        ui->graphicsView->setDragMode(QGraphicsView::NoDrag);
        ui->graphicsView->setMouseEventsEnabled(false);
        ui->graphicsView->setCurrentPointVisible(false);
        on_keyframeSpinBox_valueChanged(ui->keyframeSpinBox->value());
        drawer->setVisible(true);

        ui->actionViewPoseSpaceConnections->setChecked(false);
        ui->actionViewPoseSpacePositions->setChecked(false);
        setSplineVisible = true;
        setSplineControlVisible = false;
        setPosePointsEnabled = false;
    }
    
    for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
         it != sequences.end(); ++it)
    {
        it->second->setSplineVisible(setSplineVisible);
        it->second->setSplineControlVisible(setSplineControlVisible);
        it->second->setPosePointsEnabled(setPosePointsEnabled);
    }
    if (recalculate) compressCurrentRep();
    ui->viewer->update();
}

void MainWindow::on_keyframeSpinBox_valueChanged(int value)
{
    if (sequences.find(currentRep) == sequences.end()) return;

    Eigen::VectorXd reconVec, cs(sequences[currentRep]->getPCAComps());
    cs.setZero();
    if (value > 0) cs(value - 1) = ui->keyframeScaleSpinBox->value();
    sequences[currentRep]->evaluatePCA(cs, reconVec);
    setMeshFromVector(reconVec);
}

void MainWindow::on_keyframeScaleSpinBox_valueChanged(double /* value */)
{
    on_keyframeSpinBox_valueChanged(ui->keyframeSpinBox->value());
}

void MainWindow::on_addPoseButton_clicked()
{
    QString fileName = QFileDialog::getOpenFileName(this, tr("New shape for pose space"));

	if (fileName.length() > 0)
	{
        OpenMesh::IO::read_mesh(target, qPrintable(fileName));

        initConvertersFromTarget();
        VertexRep    vRep(*vConverter, target);
        DefGradRep   dRep(*dConverter, target);
        LinRotInvRep lRep(*lConverter, target);

        if (sequences.find(currentRep) == sequences.end())
        {
            sequences.insert(std::make_pair(Vertex, new Sequence(vRep.getDescription(), 1, parameters)));
            sequences.insert(std::make_pair(DefGrad, new Sequence(dRep.getDescription(), 1, parameters)));
            sequences.insert(std::make_pair(LinRotInv, new Sequence(lRep.getDescription(), 1, parameters)));
            
            connectSequences();
        }
        else
        {
            for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
                 it != sequences.end(); ++it)
            {
                it->second->setNumFrames(it->second->getNumFrames() + 1);
            }
        }

        sequences[Vertex]->addFrame(vRep, sequences[Vertex]->getNumFrames() - 1);
        sequences[DefGrad]->addFrame(dRep, sequences[DefGrad]->getNumFrames() - 1);
        sequences[LinRotInv]->addFrame(lRep, sequences[LinRotInv]->getNumFrames() - 1);

        for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
             it != sequences.end(); ++it)
        {
            it->second->finalize();
        }

        ++toFrame;
        compressCurrentRep();
	}
}

void MainWindow::on_unpinSelectedButton_clicked()
{
    if (sequences.find(currentRep) == sequences.end()) return;
    sequences[currentRep]->unpinSelected();
}

void MainWindow::on_unpinAllButton_clicked()
{
    if (sequences.find(currentRep) == sequences.end()) return;
    sequences[currentRep]->unpinAll();
}

void MainWindow::advanceFrame(void)
{
    double time = ui->timeSlider->value();

    if (time < ui->timeSlider->maximum())
    {
        ui->timeSlider->setNextIntegerValue();
    }
    else
    {
        stopPlaying();
        ui->timeSlider->setValue(0);
    }
}

void MainWindow::startPlaying(void)
{
    playing = true;
    ui->playStopButton->setText("Stop");
    animTimer->start(80);
}

void MainWindow::stopPlaying(void)
{
    playing = false;
    ui->playStopButton->setText("Play");
    animTimer->stop();
}

void MainWindow::initConvertersFromTarget()
{
    if (NULL != vConverter) delete vConverter;
    if (NULL != dConverter) delete dConverter;
    if (NULL != lConverter) delete lConverter;

    vConverter = new VertexRepConverter();
    dConverter = new DefGradRepConverter(target);
    lConverter = new LinRotInvRepConverter(target);
}

void MainWindow::connectSequences()
{
    for (std::map<RepOption, Sequence *>::iterator it = sequences.begin();
         it != sequences.end(); ++it)
    {
        connect(it->second, SIGNAL(newSplineCurve()), this, SLOT(splineCurveChanged()));
        it->second->viewPoseSpaceConnections(ui->actionViewPoseSpaceConnections->isChecked());
        it->second->viewPoseSpacePositions(ui->actionViewPoseSpacePositions->isChecked());
        it->second->setSplineVisible(ui->tabs->currentWidget() != ui->defineMapTab);
        it->second->fitSplineToData(ui->actionFitSpline->isChecked());
    }
}

void MainWindow::setCheckedRep() const
{
    ui->actionVertex->setChecked(currentRep == Vertex);
    ui->actionDefGrad->setChecked(currentRep == DefGrad);
    ui->actionLinRotInv->setChecked(currentRep == LinRotInv);
}

void MainWindow::loadSequence(QString fileNameRoot,
                              int fromFrame, int toFrame,
                              int numberWidth, QString extension,
                              bool createSequences)
{
    bool errorCondition = false;
    int nFrames = toFrame - fromFrame + 1;
    this->fromFrame = fromFrame;
    this->toFrame = toFrame;

    QProgressDialog progress("Loading sequence...", QString(), fromFrame, toFrame + 1, this);
    progress.setWindowTitle("Load");
    progress.setWindowModality(Qt::WindowModal);
    drawer->setVisible(false);
    
    for (int i = fromFrame; i <= toFrame; ++i)
    {
        QString fileName = QString("%1%2%3").arg(fileNameRoot).arg(QString::number(i), numberWidth, '0').arg(extension);
        progress.setValue(i);

        if (!QFile::exists(fileName))
        {
            errorCondition = true;
            break;
        }

        OpenMesh::IO::read_mesh(target, qPrintable(fileName));

        if (i == fromFrame)
        {
            initConvertersFromTarget();
        }

        VertexRep    vRep(*vConverter, target);
        DefGradRep   dRep(*dConverter, target);
        LinRotInvRep lRep(*lConverter, target);

        if (i == fromFrame)
        {
            if (createSequences)
            {
                sequences.insert(std::make_pair(Vertex, new Sequence(vRep.getDescription(), nFrames, parameters)));
                sequences.insert(std::make_pair(DefGrad, new Sequence(dRep.getDescription(), nFrames, parameters)));
                sequences.insert(std::make_pair(LinRotInv, new Sequence(lRep.getDescription(), nFrames, parameters)));
            
                connectSequences();
            }
            else
            {
                sequences[Vertex]->switchToNewDescription(vRep.getDescription());
                sequences[DefGrad]->switchToNewDescription(dRep.getDescription());
                sequences[LinRotInv]->switchToNewDescription(lRep.getDescription());
            }
        }

        if (createSequences)
        {
            sequences[Vertex]->addFrame(vRep, i - fromFrame);
            sequences[DefGrad]->addFrame(dRep, i - fromFrame);
            sequences[LinRotInv]->addFrame(lRep, i - fromFrame);
        }
        else
        {
            sequences[Vertex]->addLSVector(i - fromFrame, vRep.getVector());
            sequences[DefGrad]->addLSVector(i - fromFrame, dRep.getVector());
            sequences[LinRotInv]->addLSVector(i - fromFrame, lRep.getVector());
        }
    }
    progress.setValue(toFrame + 1);

    if (!errorCondition)
    {
        sequences[Vertex]->finalize();
        sequences[DefGrad]->finalize();
        sequences[LinRotInv]->finalize();

        compressCurrentRep();
        on_tabs_currentChanged(-1);
    }
}

void MainWindow::compressCurrentRep()
{
    if (sequences.find(currentRep) == sequences.end()) return;

    sequences[currentRep]->computeAll();
    ui->keyframeSpinBox->setMaximum(sequences[currentRep]->getPCAComps());
    ui->timeSlider->setMinimum(0);
    ui->timeSlider->setMaximum(sequences[currentRep]->getSpline().maxOfDomain());
    
    if (!statusLabelsAdded)
    {
        ui->statusBar->addPermanentWidget(&sourceStatus);
        ui->statusBar->addPermanentWidget(&pcaStatus);
        ui->statusBar->addPermanentWidget(&thinPlateStatus);
        ui->statusBar->addPermanentWidget(&splineStatus);
        statusLabelsAdded = true;
    }
    sourceStatus.setText(QString("Source: %1 x %2 ").arg(sequences[currentRep]->getNumFrames()).arg(sequences[currentRep]->getSourceRows()));
    pcaStatus.setText(QString("Shape signal: (1 + %1) x %2 ").arg(sequences[currentRep]->getPCAComps()).arg(sequences[currentRep]->getSourceRows()));
    thinPlateStatus.setText(QString("Pose signal: (3 + %1) x %2 ").arg(sequences[currentRep]->getTPSSites()).arg(sequences[currentRep]->getPCAComps()));
    splineStatus.setText(QString("Time signal: %1 x 2 ").arg(sequences[currentRep]->getNumSplinePoints()));
    
    ui->graphicsView->setScene(sequences[currentRep]->getGraphicsScene());
    ui->addPoseButton->setEnabled(true);
    ui->unpinSelectedButton->setEnabled(!sequences[currentRep]->getEmbedder().iteratingNow());
    ui->unpinAllButton->setEnabled(!sequences[currentRep]->getEmbedder().iteratingNow());
    ui->snapToCurveCheckBox->setEnabled(true);
    ui->keyframeLabel->setEnabled(true);
    ui->keyframeSpinBox->setEnabled(true);
    ui->keyframeScaleLabel->setEnabled(true);
    ui->keyframeScaleSpinBox->setEnabled(true);
    drawer->setMesh(&target);
    if (!ui->actionFreeze3DView->isChecked()) ui->viewer->showEntireScene();
    ui->graphicsView->fitInView(sequences[currentRep]->getGraphicsScene()->sceneRect(), Qt::KeepAspectRatio);
}

void MainWindow::setDialogFromParameters(Ui::ParametersDialog &dialog) const
{
    std::map<RepOption, Sequence *>::const_iterator it = sequences.find(currentRep);
    if (it != sequences.end())
    {
        int nframes = it->second->getNumFrames();
        dialog.pcaDOFSpin->setMaximum(nframes - 1);
        dialog.tpsDOFSpin->setMaximum(nframes);
        dialog.splDOFSpin->setMaximum(nframes + 2);
    }
    switch (parameters.PCATol().getType())
    {
    case Parameters::Tolerance::DegFreedom:
        dialog.pcaDOFRadio->setChecked(true);
        dialog.pcaDOFSpin->setValue(parameters.PCATol().getDegreesFreedom());
        break;
    case Parameters::Tolerance::MaxRelError:
        dialog.pcaRelErrRadio->setChecked(true);
        dialog.pcaRelErrSpin->setValue(parameters.PCATol().getMaxRelError());
        break;
    }
    switch (parameters.TPSTol().getType())
    {
    case Parameters::Tolerance::DegFreedom:
        dialog.tpsDOFRadio->setChecked(true);
        dialog.tpsDOFSpin->setValue(parameters.TPSTol().getDegreesFreedom());
        break;
    case Parameters::Tolerance::MaxRelError:
        dialog.tpsRelErrRadio->setChecked(true);
        dialog.tpsRelErrSpin->setValue(parameters.TPSTol().getMaxRelError());
        break;
    }
    switch (parameters.SPLTol().getType())
    {
    case Parameters::Tolerance::DegFreedom:
        dialog.splDOFRadio->setChecked(true);
        dialog.splDOFSpin->setValue(parameters.SPLTol().getDegreesFreedom());
        break;
    case Parameters::Tolerance::MaxRelError:
        dialog.splRelErrRadio->setChecked(true);
        dialog.splRelErrSpin->setValue(parameters.SPLTol().getMaxRelError());
        break;
    }
    dialog.tpsSmoothnessSpin->setValue(parameters.TPSSmoothness());
    dialog.splParamExpSpin->setValue(parameters.SPLAlpha());
}

void MainWindow::setParametersFromDialog(const Ui::ParametersDialog &dialog)
{
    Parameters::Tolerance pcaTol;
    if (dialog.pcaDOFRadio->isChecked()) pcaTol.setDegFreedom(dialog.pcaDOFSpin->value());
    if (dialog.pcaRelErrRadio->isChecked()) pcaTol.setRelError(dialog.pcaRelErrSpin->value());
    parameters.setPCATol(pcaTol);

    Parameters::Tolerance tpsTol;
    if (dialog.tpsDOFRadio->isChecked()) tpsTol.setDegFreedom(dialog.tpsDOFSpin->value());
    if (dialog.tpsRelErrRadio->isChecked()) tpsTol.setRelError(dialog.tpsRelErrSpin->value());
    parameters.setTPSTol(tpsTol);
    parameters.setTPSSmoothness(dialog.tpsSmoothnessSpin->value());

    Parameters::Tolerance splTol;
    if (dialog.splDOFRadio->isChecked()) splTol.setDegFreedom(dialog.splDOFSpin->value());
    if (dialog.splRelErrRadio->isChecked()) splTol.setRelError(dialog.splRelErrSpin->value());
    parameters.setSPLTol(splTol);
    parameters.setSPLAlpha(dialog.splParamExpSpin->value());
}
