//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "MonotoneInterpolator.h"
#include <cassert>
#include <algorithm>

void MonotoneInterpolator::clear()
{
    xs.clear();
    fs.clear();
}

void MonotoneInterpolator::addDataValue(double x, double f)
{
    xs.push_back(x);
    fs.push_back(f);
}

void MonotoneInterpolator::addFrozenFrames(int afterFrame, int nFrames, double value)
{
    int x = static_cast<int>(xs.back());
    for (int i = 0; i < nFrames; ++i)
    {
        xs.push_back(++x);
    }

    fs.insert(fs.begin() + afterFrame + 1, nFrames, value);
}

void MonotoneInterpolator::computeInterpolant()
{
    int n = xs.size();
    assert(n == fs.size());
    
    // Slopes
    std::vector<double> slopes(n - 1, 0);
    for (int i = 0; i < n - 1; ++i)
    {
        slopes[i] = (fs[i + 1] - fs[i]) / (xs[i + 1] - xs[i]);
    }

    // Initialize derivatives
    derivs.assign(n, 0);
    for (int i = 1; i < n - 1; ++i)
    {
        derivs[i] = 0.5 * (slopes[i - 1] + slopes[i]);
    }
    derivs[0]     = slopes[0];
    derivs[n - 1] = slopes[n - 2];
    for (int i = 1; i < n - 1; ++i)
    {
        if (fs[i + 1] == fs[i])
        {
            derivs[i] = 0;
            derivs[i + 1] = 0;
        }
    }

    // Correct derivatives
    for (int i = 0; i < n - 1; ++i)
    {
        double alpha = derivs[i] / slopes[i];
        double beta  = derivs[i + 1] / slopes[i];

        if (alpha * alpha + beta * beta > 9)
        {
            double tau = 3 / sqrt(alpha * alpha + beta * beta);
            derivs[i]     *= tau;
            derivs[i + 1] *= tau;
        }

        if (derivs[i] < 0) derivs[i] = 0;
    }
}

double MonotoneInterpolator::evaluateInterpolant(double x) const
{
    unsigned int section = findSection(x);

    double h = xs[section + 1] - xs[section];
    double t = (x - xs[section]) / h;

    double ret = 0;
    ret +=         fs[section    ] * (1 + 2 * t) * (1 - t) * (1 - t);
    ret += h * derivs[section    ] * t * (1 - t) * (1 - t);
    ret +=         fs[section + 1] * (3 - 2 * t) * t * t;
    ret += h * derivs[section + 1] * t * t * (t - 1);
    
    return ret;
}

double MonotoneInterpolator::evaluateDerivative(double x) const
{
    unsigned int section = findSection(x);

    double h = xs[section + 1] - xs[section];
    double t = (x - xs[section]) / h;

    double ret = 0;
    ret += (fs[section + 1] - fs[section]) * 6 * t * (1 - t) / h;
    ret += derivs[section    ] * (1 - t) * (1 - 3 * t);
    ret += derivs[section + 1] * (3 * t - 2) * t;
    
    return ret;
}

unsigned int MonotoneInterpolator::findSection(double x) const
{
    unsigned int section = find_if(xs.begin(), xs.end(), std::bind2nd(std::greater<double>(), x)) - xs.begin() - 1;
    if (section > xs.size() - 2) section -= 1;
    return section;
}
