//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include <QGLViewer/qglviewer.h>
#include "OpenMeshDrawer.h"

OpenMeshDrawer::OpenMeshDrawer(QGLViewer &viewer, Mesh *newMesh) :
    viewer(viewer), mesh(NULL), visible(true)
{
    connect(&viewer, SIGNAL(viewerInitialized()), this, SLOT(init()));
    connect(&viewer, SIGNAL(drawNeeded()), this, SLOT(draw()));
    setMesh(newMesh);
}

OpenMeshDrawer::~OpenMeshDrawer()
{
    if (mesh)
    {
        mesh->release_face_normals();
        mesh->release_vertex_normals();
    }
}

void OpenMeshDrawer::setMesh(Mesh *newMesh)
{
    if (mesh)
    {
        mesh->release_face_normals();
        mesh->release_vertex_normals();
    }
    mesh = newMesh;
    if (mesh)
    {
        mesh->request_face_normals();
        mesh->request_vertex_normals();
    }
    meshChanged();
}

void OpenMeshDrawer::init()
{
    viewer.setBackgroundColor(Qt::white);

    program.addShaderFromSourceFile(QGLShader::Vertex  , ":/shaders/phong.vsh");
    program.addShaderFromSourceFile(QGLShader::Fragment, ":/shaders/phong.fsh");
    bool success = program.link() && program.bind();
    if (!success)
    {
        std::cout << program.log().toStdString() << std::endl;
    }
    
    const GLfloat ambient[4]  = {0.1f, 0.1f, 0.2f, 1.0f};
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

    const GLfloat light0_position[4] = {-100.0f, 50.0f, 50.0f, 1.0f};
    const GLfloat light0_ambient[4]  = {0.0f, 0.0f, 0.0f, 1.0f};
    const GLfloat light0_diffuse[4]  = {0.6f, 0.1f, 0.1f, 1.0f};
    const GLfloat light0_specular[4] = {1.0f, 1.0f, 1.0f, 1.0f};
    glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
    glLightfv(GL_LIGHT0, GL_AMBIENT,  light0_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);

    const GLfloat light1_position[4] = {100.0f, 50.0f, 50.0f, 1.0f};
    const GLfloat light1_ambient[4]  = {0.0f, 0.0f, 0.0f, 1.0f};
    const GLfloat light1_diffuse[4]  = {0.1f, 0.1f, 0.6f, 1.0f};
    const GLfloat light1_specular[4] = {0.0f, 0.0f, 0.0f, 1.0f};
    glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
    glLightfv(GL_LIGHT1, GL_AMBIENT,  light1_ambient);
    glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_diffuse);
    glLightfv(GL_LIGHT1, GL_SPECULAR, light1_specular);
}

void OpenMeshDrawer::draw()
{
    if (mesh && visible)
    {
        glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 50.0f);
        const GLfloat material_ambient[4]  = {0.3f, 0.3f, 0.3f, 1.0f};
        const GLfloat material_diffuse[4]  = {0.9f, 0.9f, 0.9f, 1.0f};
        const GLfloat material_specular[4] = {0.9f, 0.9f, 0.9f, 1.0f};
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,  material_ambient);
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  material_diffuse);
        glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material_specular);

        glEnableClientState(GL_VERTEX_ARRAY);
        glVertexPointer(3, GL_FLOAT, 0, mesh->points());

        glEnableClientState(GL_NORMAL_ARRAY);
        glNormalPointer(GL_FLOAT, 0, mesh->vertex_normals());

        glBegin(GL_TRIANGLES);
        for (Mesh::ConstFaceIter f_it = mesh->faces_begin();
             f_it != mesh->faces_end(); ++f_it)
        {
            Mesh::ConstFaceVertexIter fv_it = mesh->cfv_iter(f_it);
            glArrayElement(fv_it.handle().idx());
            ++fv_it;
            glArrayElement(fv_it.handle().idx());
            ++fv_it;
            glArrayElement(fv_it.handle().idx());
        }
        glEnd();

        glDisableClientState(GL_VERTEX_ARRAY);
        glDisableClientState(GL_NORMAL_ARRAY);
    }
}

void OpenMeshDrawer::meshChanged()
{
    if (mesh)
    {
        mesh->update_normals();

        Mesh::Point baryCenter(0, 0, 0);
        for (Mesh::ConstVertexIter v_it = mesh->vertices_begin();
             v_it != mesh->vertices_end(); ++v_it)
        {
            baryCenter += mesh->point(v_it);
        }
        baryCenter /= mesh->n_vertices();
        viewer.setSceneCenter(qglviewer::Vec(baryCenter[0], baryCenter[1], baryCenter[2]));

        float radius = 0;
        for (Mesh::ConstVertexIter v_it = mesh->vertices_begin();
             v_it != mesh->vertices_end(); ++v_it)
        {
            radius = qMax((mesh->point(v_it) - baryCenter).norm(), radius);
        }
        if (radius > 0) viewer.setSceneRadius(radius);
    }

    viewer.update();
}
