//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "SubspaceRecon.h"
#include "Sequence.h"

#include <Eigen/SVD>
#include <iostream>
#include <QElapsedTimer>
#include <QProgressDialog>

SubspaceRecon::SubspaceRecon(const Sequence &sequence)
    : sequence(sequence), useComps(-1),
    offset(Eigen::VectorXd::Zero(sequence.getSourceRows()))
{}

void SubspaceRecon::initSubspace(int nComps)
{
    useComps = nComps;
    subspace = Eigen::MatrixXd::Zero(sequence.getSourceRows(), useComps);
}

void SubspaceRecon::setOffset(const Eigen::VectorXd &newOffset)
{
    offset = newOffset;
}

void SubspaceRecon::setSubspaceVector(int comp, const Eigen::VectorXd &vector)
{
    subspace.col(comp) = vector - offset;
}

void SubspaceRecon::computeLSCoeffs(const Eigen::MatrixXd &normalizedFeatures)
{
    Eigen::MatrixXd data = normalizedFeatures;
    data.colwise() -= offset;
    QElapsedTimer timer;
    timer.start();
    coeffs = Eigen::JacobiSVD<Eigen::MatrixXd>(subspace, Eigen::ComputeThinU | Eigen::ComputeThinV).solve(data);
    std::cout << "Calculating LSC took " << timer.elapsed() << "ms" << std::endl;
}

void SubspaceRecon::resetSubspace()
{
    offset = Eigen::VectorXd::Zero(sequence.getSourceRows());
    subspace = Eigen::MatrixXd::Zero(sequence.getSourceRows(), useComps);
}

void SubspaceRecon::computePCA(const Eigen::MatrixXd &normalizedFeatures,
                               const Parameters::Tolerance &tol,
                               QProgressDialog *progress)
{
    Eigen::MatrixXd data = normalizedFeatures;
    offset = data.rowwise().sum();
    offset /= data.cols();
    data.colwise() -= offset;
    QElapsedTimer timer;
    timer.start();
    Eigen::JacobiSVD<Eigen::MatrixXd> svd(data.transpose(), Eigen::ComputeThinV);
    std::cout << "Calculating PCA took " << timer.elapsed() << "ms" << std::endl;
    subspace = svd.matrixV();
    coeffs = svd.matrixV().transpose() * data;
    
    switch (tol.getType())
    {
    case Parameters::Tolerance::DegFreedom:
        {
            int dof = qMin(tol.getDegreesFreedom(), sequence.getNumFrames());
            useComps = dof;
        }
        break;
    case Parameters::Tolerance::MaxRelError:
        for (int nComps = 2; nComps < sequence.getNumFrames(); ++nComps)
        {
            if (progress) progress->setValue(progress->value() + 1);
            Eigen::MatrixXd recon;
            reconstructAllFrames(recon, nComps);
            Eigen::VectorXd errors;
            sequence.denormalizeAndRelErr(recon, errors);

            if (errors.maxCoeff() < tol.getMaxRelError())
            {
                useComps = nComps;
                break;
            }
        }
        break;
    }
    if (useComps < 0) useComps = sequence.getNumFrames() - 1;
}

void SubspaceRecon::reconstructFrame(int col, Eigen::VectorXd &result, int nComps) const
{
    if (nComps < 0) nComps = useComps;
    result = subspace.leftCols(nComps) * coeffs.block(0, col, nComps, 1);
    result += offset;
}

void SubspaceRecon::reconstructAllFrames(Eigen::MatrixXd &result, int nComps) const
{
    if (nComps < 0) nComps = useComps;
    result = subspace.leftCols(nComps) * coeffs.topRows(nComps);
    result.colwise() += offset;
}

void SubspaceRecon::reconstructFrom(const Eigen::VectorXd &cs, Eigen::VectorXd &result) const
{
    result = subspace.leftCols(useComps) * cs;
    result += offset;
}

void SubspaceRecon::reconstructFrom(const Eigen::MatrixXd &cs, Eigen::MatrixXd &result) const
{
    result = subspace.leftCols(useComps) * cs;
    result.colwise() += offset;
}

void SubspaceRecon::getCoeffSubMatrix(Eigen::MatrixXd &result) const
{
    result = coeffs.topRows(useComps);
}

