//
//    Copyright (c) 2011 Thomas J. Cashman and Kai Hormann
//                       Tom.Cashman@cantab.net / Kai.Hormann@usi.ch
//
//    This file is part of DefMeshRep.
//
//    DefMeshRep is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    DefMeshRep is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with DefMeshRep.  If not, see <http://www.gnu.org/licenses/>.
//

#include "ThinPlateRecon.h"
#include "Sequence.h"
#include "EmbedderMDS.h"
#include <Eigen/QR>
#include <Eigen/SVD>
#include <QElapsedTimer>
#include <QProgressDialog>
#include <iostream>
#include <numeric>

ThinPlateRecon::ThinPlateRecon(const Parameters::Tolerance &tol,
                               const Sequence &sequence,
                               bool haveGroundTruth,
                               double smoothnessParameter,
                               QProgressDialog *progress) :
    tol(tol), sequence(sequence),
    p(smoothnessParameter),
    X(sequence.getNumFrames(), 3),
    coeffs(sequence.getPCAComps(), sequence.getNumFrames()),
    linearPart(sequence.getPCAComps(), 3)
{
    int nframes = sequence.getNumFrames();
    sequence.getSubspace().getCoeffSubMatrix(pcaCoeffs);
    for (int i = 0; i < nframes; ++i)
    {
        X(i, 0) = sequence.getEmbedder().frameX(i);
        X(i, 1) = sequence.getEmbedder().frameY(i);
        X(i, 2) = 1;
    }

    QElapsedTimer timer;
    timer.start();

    Eigen::MatrixXd q2Full = X.colPivHouseholderQr().householderQ();
    q2 = q2Full.rightCols(nframes - 3);

    // First four points from max error on linear fit
    included.assign(nframes, false);
    Eigen::Matrix3d A = Eigen::Matrix3d::Zero();
    Eigen::MatrixXd B(3, pcaCoeffs.rows());
    B.setZero();
    for (int i = 0; i < nframes; ++i)
    {
        A += X.row(i).transpose() * X.row(i);
        B += X.row(i).transpose() * pcaCoeffs.col(i).transpose();
    }
    linearPart = A.ldlt().solve(B).transpose();
    
    Eigen::VectorXd errors;
    if (haveGroundTruth)
    {
        Eigen::MatrixXd recon;
        sequence.getSubspace().reconstructFrom(linearPart * X.transpose(), recon);
        sequence.denormalizeAndRelErr(recon, errors);
    }
    else
    {
        Eigen::MatrixXd recon = linearPart * X.transpose();
        errors = (recon - pcaCoeffs).colwise().norm().array() / pcaCoeffs.colwise().norm().array();
        for (int i = 0; i < errors.size(); ++i) errors[i] = Sequence::validateError(errors[i]);
    }
    std::vector<int> indices(nframes);
    for (int i = 0; i < nframes; ++i) indices[i] = i;
    partial_sort(indices.begin(), indices.begin() + 4, indices.end(), index_compare_greater(errors.data()));
    for (int i = 0; i < 4; ++i)
    {
        included[indices[i]] = true;
    }

    while (true)
    {
        calculate();
        if (progress) progress->setValue(progress->value() + 1);
        if (numIncludedSites() == nframes) break;
        if (tol.getType() == Parameters::Tolerance::DegFreedom &&
            numIncludedSites() >= tol.getDegreesFreedom()) break;
        
        Eigen::VectorXd errors;
        Eigen::MatrixXd pcaCoeffRecon(pcaCoeffs.rows(), pcaCoeffs.cols());
        for (int i = 0; i < nframes; ++i)
        {
            Eigen::VectorXd temp;
            evaluate(sequence.getEmbedder().frameX(i), sequence.getEmbedder().frameY(i), temp);
            pcaCoeffRecon.col(i) = temp;
        }
        if (haveGroundTruth)
        {
            Eigen::MatrixXd recon;
            sequence.getSubspace().reconstructFrom(pcaCoeffRecon, recon);
            sequence.denormalizeAndRelErr(recon, errors);
        }
        else
        {
            errors = (pcaCoeffRecon - pcaCoeffs).colwise().norm().array() / pcaCoeffs.colwise().norm().array();
            for (int i = 0; i < errors.size(); ++i) errors[i] = Sequence::validateError(errors[i]);
        }

        if (tol.getType() == Parameters::Tolerance::MaxRelError &&
            *std::max_element(errors.data(), errors.data() + nframes) < tol.getMaxRelError()) break;
            
        int addSite;
        do
        {
            addSite = std::max_element(errors.data(), errors.data() + nframes) - errors.data();
            errors[addSite] = 0;
        } while (included[addSite]);
        included[addSite] = true;
    }

    std::cout << "Total time for TPS:  " << timer.elapsed() << "ms" << std::endl;
 }

void ThinPlateRecon::calculate()
{
    QElapsedTimer timer;
    timer.start();
    int nsites  = numIncludedSites();
    int nframes = sequence.getNumFrames();
    Eigen::MatrixXd X_subset(nsites, 3);
    
    int coeffIx = 0;
    for (int i = 0; i < nframes; ++i)
    {
        if (!included[i]) continue;

        X_subset(coeffIx, 0) = X(i, 0);
        X_subset(coeffIx, 1) = X(i, 1);
        X_subset(coeffIx, 2) = 1;
        ++coeffIx;
    }

    Eigen::ColPivHouseholderQR<Eigen::MatrixXd> qr_subset = X_subset.colPivHouseholderQr();
    const Eigen::VectorXd rDiag = qr_subset.matrixQR().diagonal().array().abs();
    if (rDiag.minCoeff() < 1e-14 * rDiag.maxCoeff())
    {
        std::cerr << "Some nontrivial linear polynomial vanishes at all data sites" << std::endl;
    }
    
    Eigen::MatrixXd q2_subset = qr_subset.householderQ();

    Eigen::MatrixXd colmat(nframes, nsites);
    int coeffIxI = 0;
    for (int i = 0; i < nframes; ++i)
    {
        int coeffIxJ = 0;
        for (int j = 0; j < nframes; ++j)
        {
            if (!included[j]) continue;

            double rsq = (X(i, 0) - X(j, 0)) * (X(i, 0) - X(j, 0)) +
                         (X(i, 1) - X(j, 1)) * (X(i, 1) - X(j, 1));

            double val = (rsq == 0 ? 0 : rsq * log(rsq));
            
            colmat(i, coeffIxJ) = val;
            ++coeffIxJ;
        }

        if (included[i])
        {
            colmat(i, coeffIxI) = (1 - p) / p;
            ++coeffIxI;
        }
    }
    
    const Eigen::MatrixXd gamma =
        Eigen::JacobiSVD<Eigen::MatrixXd>(q2.transpose() * colmat * q2_subset.rightCols(nsites - 3),
                                          Eigen::ComputeThinU | Eigen::ComputeThinV)
                                          .solve(q2.transpose() * pcaCoeffs.transpose());
    coeffs = (q2_subset.rightCols(nsites - 3) * gamma).transpose();
    
    const Eigen::MatrixXd solveFor = (pcaCoeffs - coeffs * colmat.transpose()).transpose();
    linearPart = Eigen::JacobiSVD<Eigen::MatrixXd>(X, Eigen::ComputeThinU | Eigen::ComputeThinV)
                 .solve(solveFor).transpose();
    
    //std::cout << "Calculating TPS took " << timer.elapsed() << "ms" << std::endl;
}

int ThinPlateRecon::numIncludedSites(void) const
{
    return std::accumulate(included.begin(), included.end(), 0);
}
    
void ThinPlateRecon::evaluate(double x, double y, Eigen::VectorXd &dest) const
{
    dest  =     linearPart.col(2);
    dest += x * linearPart.col(0);
    dest += y * linearPart.col(1);
    
    int coeffIx = 0;
    for (int i = 0; i < sequence.getNumFrames(); ++i)
    {
        if (!included[i]) continue;

        double dx  = x - X(i, 0);
        double dy  = y - X(i, 1);
        double rsq = dx * dx + dy * dy;
        double val = (rsq == 0 ? 0 : rsq * log(rsq));
        
        dest += val * coeffs.col(coeffIx);
        ++coeffIx;
    }
}
